"""
Luis Alvarez Pattern (New)
"""

from lib import base_app, build, image, http
from lib.misc import ctime
from lib.base_app import init_app
import cherrypy
from cherrypy import TimeoutError
import os.path
import time
import shutil

#
# INTERACTION
#

class NoMatchError(RuntimeError):
    """ Exception raised when ... """
    pass

class app(base_app):
    """ Template demo app """
    
    title = "Luis Alvarez Pattern (New)"

    input_nb = 2 # number of input images
    input_max_pixels = 512*512 # max size (in pixels) of an input image
    input_max_method = 'zoom'
    input_dtype = '3x8i' # input image expected data type    
    input_ext = '.png'   # input image expected extension (ie file format)    
    is_test = False      # switch to False for deployment

    is_listed = True
    is_built = True
    id_archive = 0

    xlink_article = "http://www.ipol.im/pub/pre/100/"
    xlink_src = "http://serdis.dis.ulpgc.es/~asalgado/image_synthesisPNG.zip"
    
    parconfig = {}
    parconfig['iwidth'] = {'type':int, 'default': 800, 'htmltype': 'text',
            'changeable': True, 'htmlname': 'Width', 'doc':'', 'value': 800}
    parconfig['iheight'] = {'type':int, 'default': 600, 'htmltype': 'text',
            'changeable': True, 'htmlname': 'Height', 'doc':'', 'value': 600}
    parconfig['params_file'] = {'type':str, 'default': 'input.info', 
            'htmltype': 'file', 'changeable': True, 'htmlname': 'Params file',
            'doc':'(optional)', 'value': 'input.info'}



    def __init__(self):
        """
        app setup
        """
        # setup the parent class
        base_dir = os.path.dirname(os.path.abspath(__file__))
        base_app.__init__(self, base_dir)



    def __build_algo__(self):
        """
        program build/update
        """
        ## store common file path in variables
        tgz_file  = self.dl_dir + "image_synthesisPNG.zip"
        prog_file = self.bin_dir + "image_synthesis"
        log_file  = self.base_dir + "build.log"
        ## get the latest source archive
        build.download(app.xlink_src, tgz_file)
        ## test if the dest file is missing, or too old
        if (os.path.isfile(prog_file)
            and ctime(tgz_file) < ctime(prog_file)):
            cherrypy.log("no rebuild needed",
                      context='BUILD', traceback=False)
        else:
            # extract the archive
            build.extract(tgz_file, self.src_dir)
            # build the program
            build.run("make -C %s" % (self.src_dir +"image_synthesisPNG"), \
                      stdout=log_file)
            # save into bin dir
            if os.path.isdir(self.bin_dir):
                shutil.rmtree(self.bin_dir)
            os.mkdir(self.bin_dir)

            shutil.copy(self.src_dir + os.path.join("image_synthesisPNG",
                        "image_synthesis"), prog_file)
            # cleanup the source dir
            shutil.rmtree(self.src_dir)
            shutil.rmtree(self.dl_dir)
        return


    def build(self):
        self.__build_algo__()
        return
    



    @cherrypy.expose
    @init_app
    def wait(self, **kwargs):
        """
        params handling and run redirection
        """
        try:
            for k in app.parconfig:
                typ = app.parconfig[k]['type']
                if k in kwargs:
                    self.cfg['param'][k] = typ(kwargs[k])
                    app.parconfig[k]['value'] = typ(kwargs[k])
                else:
                    self.cfg['param'][k] = typ('') 
                    app.parconfig[k]['value'] = typ('') 
                   
            if 'params_file' in kwargs:
                typ = app.parconfig['params_file']['type']
                k = 'params_file' 
            
                self.cfg['param'][k] = typ('') 
                app.parconfig[k]['value'] = typ('')

                if 'run' in kwargs:
                    #if kwargs['run'] == "restart with the same palette":
                        #self.cfg['param'][k] = typ(kwargs[k])
                        #app.parconfig[k]['value'] = typ(kwargs[k])
                        ##print("RESTART PALETTE = "
                        ##+ str(kwargs['params_file']))

                    if kwargs['run'] == "change":
                        if self.upload_given_file('input.info', kwargs[k]) :
                            self.cfg['param'][k] = typ(kwargs[k])
                            app.parconfig[k]['value'] = typ(kwargs[k])
                else:
                    if self.upload_given_file('input.info', kwargs[k]) :
                        self.cfg['param'][k] = typ(kwargs[k])
                        app.parconfig[k]['value'] = typ(kwargs[k])

        except ValueError:
            return self.error(errcode='badparams',
                     errmsg='The parameters must be numeric.')


        http.refresh(self.base_url + 'run?key=%s' % self.key)
        self.cfg['meta']['height'] = app.parconfig['iheight']['default']
        #image(self.work_dir + 'input_0.png').size[1]
        self.cfg['meta']['width']  = app.parconfig['iwidth']['default']
        #image(self.work_dir + 'input_0.png').size[0]
        self.cfg.save()
        
        return self.tmpl_out("wait.html",
                             height=app.parconfig['iheight']['default'])
                             #image(self.work_dir + 'input_0.png').size[1])




    @cherrypy.expose
    @init_app
    def run(self, **kwargs):
        """
        algo execution
        """
        try:
            run_time = time.time()
            self.__run_algo__(timeout=None)
            self.cfg['info']['run_time'] = time.time() - run_time
        except TimeoutError:
            return self.error(errcode='timeout',
                              errmsg="Try again with simpler images.")
        except NoMatchError:
            http.redir_303(self.base_url +
                           'result?key=%s&error_nomatch=1' % self.key)
        except RuntimeError:
            return self.error(errcode='runtime')
        else:
            http.redir_303(self.base_url + 'result?key=%s' % self.key)

        return self.tmpl_out("run.html")
    

    @cherrypy.expose
    @init_app
    def save_archive(self, **kwargs):
        """
        save archive
        """
        try:
            app_archive = self.make_archive()
            app_archive.add_file("input_0.png", \
                                 "palette_" + str(self.id_archive) + ".png", \
                                 info="palette")
            app_archive.add_file("input_1.png", \
                                 "silhoette_" + str(self.id_archive)+ ".png", \
                                 info="silhoette")
            if self.cfg['param']['params_file'] != '' :
                app_archive.add_file("input.info", "input_" \
                                    + str(self.id_archive) + ".info", \
                                    info="input parameters file")
            app_archive.add_file("output.png", "output_" \
                                    + str(self.id_archive) + ".png", \
                                    info="output image")
            app_archive.add_file("output.info", "output_" \
                                    + str(self.id_archive) + ".info", \
                                    info="output parameters file")
            app_archive.add_info({"width" + str(self.id_archive): \
                                           self.cfg['param']['iwidth'],
                                  "height" + str(self.id_archive): \
                                           self.cfg['param']['iheight']
                                 })
            app_archive.save()

            self.id_archive = self.id_archive + 1 

        except RuntimeError:
            return self.error(errcode='runtime')
        http.redir_303(self.base_url + 'result?key=%s' % self.key)

        return self.tmpl_out("run.html")
    
    


    def __run_algo__(self, timeout=None):
        """
        the core algo runner
        could also be called by a batch processor
        this one needs no parameter
        """
        width              = self.cfg['param']['iwidth']
        height             = self.cfg['param']['iheight']
        palette            = self.work_dir + "input_0.png"

        silhouette         = self.work_dir + "input_1.png"
        if not os.path.exists( silhouette ):
            silhouette = "null.png"

        input_params_file  = ''
        if self.cfg['param']['params_file'] != '' :
            input_params_file  = self.work_dir + "input.info"

        output_image       = self.work_dir + "output.png"
        output_params_file = self.work_dir + "output.info"


        if self.cfg['param']['params_file'] != '' :      
            my_proc = self.run_proc(['image_synthesis',
                      str(silhouette),
                      str(width),
                      str(height),
                      str(output_image),
                      str(output_params_file),
                      str(palette),
                      str(input_params_file)
                     ])
        else:
            my_proc = self.run_proc(['image_synthesis',
                      str(silhouette),
                      str(width),
                      str(height),
                      str(output_image),
                      str(output_params_file),
                      str(palette)
                     ])

        self.wait_proc(my_proc, timeout=self.timeout)
        return




    @cherrypy.expose
    @init_app
    def result(self, public=None):
        """
        display the algo results
        """
        return self.tmpl_out("result.html",
                              height=app.parconfig['iheight']['default'])

    
    
    
    def upload_given_file(self, i, fich):
        """
        upload a file
        """
        file_up = fich
        if '' == file_up.filename:
            # missing file
            return False
        file_save = file(self.work_dir + i, 'wb')
        size = 0
        while True:
            # T O D O larger data size
            data = file_up.file.read(128)
            if not data:
                break
            size += len(data)
            if size > self.input_max_weight:
                # file too heavy
                raise cherrypy.HTTPError(400, # Bad Request
                        "File too large, " +
                        "resize or use better compression")
            file_save.write(data)
        file_save.close()
        return True
    
    

    @cherrypy.expose
    def input_upload(self, **kwargs):
        """
        use the uploaded input images
        """
        self.new_key()
        self.init_cfg()
        msg = "no message"
        
        self.log("input uploaded")
        self.cfg['meta']['original'] = True

        ## .............................................

        rfa = self.upload_given_file('input_0', kwargs['file_0'])
        if not rfa:
            return self.error(errcode='badparams',
                         errmsg='(MISSING PALETTE)')

        try:
            ima = image(self.work_dir + 'input_0')
        except IOError:
            return self.error(errcode='badparams',
                     errmsg='(INPUT IMAGE FORMAT ERROR)')

        if self.input_max_pixels and \
           (ima.size[0]*ima.size[1]) > (self.input_max_pixels):
            ima.resize(self.input_max_pixels)
            msg = """The image has been resized
                  for a reduced computation time."""

        self.cfg['meta']['height'] = ima.size[1]
        self.cfg['meta']['width']  = ima.size[0]

        ima.save(self.work_dir + 'input_0.png')


        ## .............................................
        
        rfb = self.upload_given_file('input_1', kwargs['file_1'])
        if rfb:

            try:
                imb = image(self.work_dir + 'input_1')
            except IOError:
                return self.error(errcode='badparams',
                         errmsg='(INPUT IMAGE FORMAT ERROR)')

            if self.input_max_pixels and \
               (imb.size[0]*imb.size[1]) > (self.input_max_pixels):
                ima.resize(self.input_max_pixels)
                msg = """The image has been resized
                      for a reduced computation time."""

            self.cfg['meta']['height'] = imb.size[1]
            self.cfg['meta']['width'] = imb.size[0]

            imb.save(self.work_dir + 'input_1.png')

        else :
            shutil.copy(self.base_dir + os.path.join("input",
                        "silhouette_white.png"),
                        self.work_dir + 'input_1.png')
        
        ## .............................................

        self.cfg.save()

        # jump to the params page
        return self.params(msg=msg, key=self.key)
